# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
import math
from dataclasses import dataclass
from typing import Optional

import torch
from torch import nn

from xformers.components.attention import Attention, AttentionConfig, register_attention

_mask_type_warning = True

_use_triton = torch.cuda.is_available()
if _use_triton:
    try:
        from triton.ops.blocksparse import matmul as blocksparse_matmul
        from triton.ops.blocksparse import softmax as blocksparse_softmax

        from xformers.triton.softmax import MaskType
        from xformers.triton.utils import gpu_capabilities_older_than_70

        # Blocksparse requires Tensor cores
        if gpu_capabilities_older_than_70():
            logging.warning(
                "Blocksparse is not available: the current GPU does not expose Tensor cores"
            )
            _use_triton = False

    except ImportError as e:
        logging.warning(
            f"Triton is not available: {e}.\nBlockSparse attention will not be available"
        )
        _use_triton = False


if _use_triton:

    @dataclass
    class BlockSparseAttentionConfig(AttentionConfig):
        layout: torch.Tensor  # The dimensions of the random features
        block_size: int
        dropout: float
        num_heads: int

    @register_attention("blocksparse", BlockSparseAttentionConfig)
    class BlockSparseAttention(Attention):
        r"""
        Thin wrap over the Triton blocksparse computations. The sparsity pattern is determined through the layout.

        .. warning: the layout is assumed to have the dimensions [heads, seq, seq].
            If some dimensions are missing, we assume that the same layout is to be used across heads.

        .. warning: for now, the sequence (context) length has to be a power of two. This constraint could
            be relaxed in the future.

        .. note: it is possible to pass a specific per batch mask in the forward call,
            but this will not lead to any speed up.
            Any constant sparsity pattern is better passed through the layout parameter.
        """

        def __init__(
            self,
            layout: torch.Tensor,
            block_size: int = 16,
            dropout: float = 0.0,
            num_heads: int = 1,  # optional, used to adapt the layout if in need
            *args,
            **kwargs,
        ):
            if layout.dim() == 2:
                logging.warning(
                    "The layout passed is lacking a head dimension and a batch dimension"
                )
                logging.warning(
                    "Now assuming that the same layout is to be used across all heads"
                )
                layout = layout.unsqueeze(0).expand(num_heads, -1, -1)
                logging.warning(f"New layout dimensions: {layout.shape}")

            assert block_size >= 16, "Minimum block size is 16, for now at least"

            super().__init__()
            self.attn_drop = nn.Dropout(dropout, inplace=False)

            # Pure blocksparse data
            self.layout = layout
            self.block_size = block_size

            # blocksparse operators
            self.sparse_dot_sdd = blocksparse_matmul(
                self.layout,
                self.block_size,
                "sdd",
                trans_a=False,
                trans_b=True,
            )
            self.sparse_dot_dsd = blocksparse_matmul(
                self.layout,
                self.block_size,
                "dsd",
                trans_a=False,
                trans_b=False,
            )
            self.sparse_softmax = blocksparse_softmax(self.layout, self.block_size)

            # make sure that the head dimension is not folded down with the batch
            self.requires_head_dimension = True

            # key padding mask and attention mask must be passed in separately
            self.requires_separate_masks = True

        def update_mask_type(self, mask: torch.Tensor, to_dtype: torch.dtype):
            global _mask_type_warning
            if _mask_type_warning:
                logging.warning(
                    "Mask has to be multiplicative. Fixing that but this slows things down"
                )
                _mask_type_warning = False  # Only warn once
            mask = mask.to(to_dtype)

        def forward(
            self,
            q: torch.Tensor,
            k: torch.Tensor,
            v: torch.Tensor,
            att_mask: Optional[torch.Tensor] = None,
            key_padding_mask: Optional[torch.Tensor] = None,
            scale: float = 1.0,
            *args,
            **kwargs,
        ) -> torch.Tensor:
            r"""
            att_mask            A 2D attention mask. The dtype must be the same as q. Multiplicative mask where a value
                                of 1 will keep the value, while a value of 0 will mask the value.
            key_padding_mask    A mask with size (batch size x sequence length). The dtype must be the same as q.
                                Multiplicative mask where a value of 1 will keep the value, while a value of 0 will
                                mask the value.
            """

            # NOTE:
            # The attention mask will be taken into account when computing the softmax
            # meaning that non-masked values which are present in the initial blocksparse layout will be computed.
            # If blocks are to be constantly masked, better perf would thus be reached by signalling them out in the
            # initial attention setup

            if att_mask is not None and att_mask.dtype != q.dtype:
                self.update_mask_type(att_mask, q.dtype)
            if key_padding_mask is not None and key_padding_mask.dtype != q.dtype:
                self.update_mask_type(key_padding_mask, q.dtype)

            assert (
                att_mask is None or att_mask.dim() == 2
            ), "The attention mask is constant across heads, expected dimensions are [seq x seq]"

            assert (
                q.shape[-2] == k.shape[-2]
            ), "Blocksparse requires the same dimensions for K and Q for now"

            assert (
                q.shape[-2] == self.layout.shape[-2] * self.block_size
            ), "Actual sequence size and layout are inconsistent"
            assert (
                k.shape[-2] == self.layout.shape[-2] * self.block_size
            ), "Actual sequence size and layout are inconsistent"

            assert math.log(
                q.shape[-2], 2
            ).is_integer(), (
                "For now blocksparse only works on power-of-two sequence lengths"
            )

            # Blocksparse only works on fp16
            q_dtype = q.dtype
            q, k, v = q.half(), k.half(), v.half()

            if att_mask is not None:
                att_mask = att_mask.half()

            if key_padding_mask is not None:
                key_padding_mask = key_padding_mask.half()

            # Self-attend: (B, nh, S, hs) x (B, nh, hs, S) -> (B, nh, S, S)
            # When the computations are block sparse, the matrix types change along the way:
            # - (sparse) attention matrix = (dense) Kt * (dense) Q
            q = q / math.sqrt(q.size(-1))
            sparse_att_mat = self.sparse_dot_sdd(q, k)

            # - softmax on the sparse attention matrix
            sparse_att_mat = self.sparse_softmax(
                sparse_att_mat,
                scale=scale,
                key_padding_mask=key_padding_mask,
                attn_mask=att_mask,
                key_padding_mask_mode=MaskType.MUL,
                attn_mask_mode=MaskType.MUL,
            )

            # - then (dense) attention is (sparse) attention matrix * dense (value)
            a = self.sparse_dot_dsd(sparse_att_mat, v)
            return a.to(q_dtype)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from abc import ABCMeta, abstractmethod
from dataclasses import asdict, dataclass
from typing import Optional, Type, TypeVar

import torch
import torch.nn as nn


@dataclass
class AttentionConfig:
    """Parameters required for all Attentions.
    Can accept and store extra parameters.
    """

    name: str  # the registered name for this attention mechanism
    dropout: float  # dropout probability


Self = TypeVar("Self", bound="Attention")


# Define the common interface, every attention block needs to derive from it
class Attention(nn.Module, metaclass=ABCMeta):
    r"""The base Attention mechanism, which is typically a sub-part of the multi-head attention"""

    _causal_mask: Optional[torch.Tensor] = None

    @abstractmethod
    def __init__(self, dropout: Optional[float] = None, *args, **kwargs):
        super().__init__()
        self.requires_input_projection = True
        self.requires_head_dimension = False
        # key padding mask and attention mask must be passed in as separate arguments instead of a merged attention mask
        self.requires_separate_masks = False

    @classmethod
    def from_config(cls: Type[Self], config: AttentionConfig) -> Self:
        # Generate the class inputs from the config
        fields = asdict(config)

        # Skip all Nones so that default values are used
        fields = {k: v for k, v in fields.items() if v is not None}

        return cls(**fields)

    @abstractmethod
    def forward(
        self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, *args, **kwargs
    ) -> torch.Tensor:
        raise NotImplementedError

    def _get_causal_mask(self, seq_len: int, to_seq_len: int) -> torch.Tensor:
        # Cache a mask so that multiple instances would reuse the same
        causal_mask = self._causal_mask
        if not causal_mask:
            causal_mask = torch.tril(torch.ones(seq_len, to_seq_len), diagonal=0)
            causal_mask[self._causal_mask == 1] = -float("inf")
            causal_mask.unsqueeze_(0)  # batch dimension
            self._causal_mask = causal_mask

        return causal_mask



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS: This comes almost as-is from the Triton dropout tutorial
# https://raw.githubusercontent.com/openai/triton/master/python/tutorials/04-low-memory-dropout.py

import torch
import triton
from torch.cuda.amp import custom_bwd, custom_fwd

from xformers.triton.k_dropout import k_dropout


# Helper to handle the SPMD launch grid and error cases
class _dropout(torch.autograd.Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(ctx, x, p):
        # Soft-flatten an hypothetical 3rd dimension
        x_ = x.reshape(-1, x.shape[-1])
        y = torch.empty_like(x_)
        _, N = x_.shape

        assert y.stride(-1) == 1 and x_.stride(-1) == 1

        # Generate one seed per sample
        # seed max is int32 max for positive numbers: 2**16
        seeds = torch.randint(65536, (x_.shape[0],), device=x.device).to(torch.int32)

        # SPMD launch grid
        def grid(meta):
            return (
                x_.shape[0],
                triton.cdiv(x_.shape[1], meta["BLOCK_SIZE"]),
            )

        k_dropout[grid](y, x_, seeds, y.stride(0), N, p)

        ctx.save_for_backward(seeds)
        ctx.p = p

        return y.reshape_as(x)

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_out):
        (seeds,) = ctx.saved_tensors

        # Soft-flatten an hypothetical 3rd dimension
        grad_out_ = grad_out.reshape(-1, grad_out.shape[-1])
        grad_in = torch.empty_like(grad_out_)
        _, N = grad_out_.shape

        assert grad_in.stride(-1) == 1 and grad_out_.stride(-1) == 1

        # SPMD launch grid
        def grid(meta):
            return (
                grad_out_.shape[0],
                triton.cdiv(grad_out_.shape[1], meta["BLOCK_SIZE"]),
            )

        k_dropout[grid](grad_in, grad_out_, seeds, grad_out_.stride(0), N, ctx.p)

        return grad_in.reshape_as(grad_out), None


def dropout(x: torch.Tensor, p: float):
    if p > 0.0:
        return _dropout.apply(x, p)

    return x



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import torch
import triton

from xformers.triton.utils import get_current_cuda_device

# CREDITS: Optimized defaults as suggested in the Triton documentation for matrix multiplications


# Handle different SM configurations for the older GPUs
# fmt: off
_configs_P100 = [
        triton.Config({'BLOCK_ROW': 64 , 'BLOCK_COL': 32}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_ROW': 32 , 'BLOCK_COL': 64}, num_stages=5, num_warps=2)
]

_configs_V100 = _configs_P100 + [
        triton.Config({'BLOCK_ROW': 128, 'BLOCK_COL': 256}, num_stages=3, num_warps=8),
        triton.Config({'BLOCK_ROW': 256, 'BLOCK_COL': 128}, num_stages=3, num_warps=8),
        triton.Config({'BLOCK_ROW': 256, 'BLOCK_COL': 64}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_ROW': 64 , 'BLOCK_COL': 256}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_ROW': 128, 'BLOCK_COL': 128}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_ROW': 128, 'BLOCK_COL': 64}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_ROW': 64 , 'BLOCK_COL': 128}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_ROW': 128, 'BLOCK_COL': 32}, num_stages=4, num_warps=4),
]

# fmt: on

# NOTE: Could be that different configs would be better
_configs_A100 = _configs_V100

kernel_config = (
    {
        "P100": _configs_P100,
        "T4": _configs_P100,
        "V100": _configs_V100,
        "A100": _configs_A100,
    }[get_current_cuda_device()]
    if torch.cuda.is_available()
    else {}
)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from typing import Optional

import torch
from torch import nn

from xformers.components.attention import Attention, AttentionConfig, register_attention
from xformers.components.attention.core import scaled_dot_product_attention


@dataclass
class ScaledDotProductConfig(AttentionConfig):
    causal: Optional[bool]
    seq_len: Optional[int]
    to_seq_len: Optional[int]


@register_attention("scaled_dot_product", ScaledDotProductConfig)
class ScaledDotProduct(Attention):
    r"""
    Implementing the Scaled Dot-Product attention proposed in
    `Attention is all you need`_, Vaswani et al.

    .. _`Attention is all you need`: https://arxiv.org/abs/1706.03762v5
    """

    mask: Optional[torch.Tensor]

    def __init__(
        self,
        dropout: float = 0.0,
        causal: bool = False,
        seq_len: Optional[int] = None,
        to_seq_len: Optional[int] = None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.attn_drop = nn.Dropout(dropout, inplace=False)
        self.causal = causal
        self.seq_len = seq_len

        if causal and seq_len is not None:
            mask = self._get_causal_mask(seq_len, to_seq_len if to_seq_len else seq_len)
            self.register_buffer("mask", mask)
        else:
            self.mask = None

    def forward(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        att_mask: Optional[torch.Tensor] = None,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        r"""
        att_mask    A 2D or 3D mask which ignores attention at certain positions.

                    - If the mask is boolean, a value of True will keep the value,
                        while a value of False will mask the value.

                        Key padding masks (dimension: batch x sequence length) and attention masks
                        (dimension: sequence length x sequence length OR batch x sequence length x sequence length)
                        can be combined and passed in here. Method maybe_merge_masks provided in the utils can be
                        used for that merging.

                    - If the mask has the float type, then an additive mask is expected (masked values are -inf)

        """
        # Mask-aware attention
        if self.mask is not None:
            self.mask = self.mask.to(q.dtype)

            if att_mask is not None:
                if att_mask.ndim == 2:
                    att_mask.unsqueeze_(0)

                if att_mask.dtype == torch.bool:
                    # bool mask
                    att_mask_ = self.mask
                    att_mask_[~att_mask] = float("-inf")
                    att_mask = att_mask_
                else:
                    # additive mask
                    att_mask = self.mask + att_mask

                att_mask = att_mask.to(q.dtype)

        # Self-attend: (B x nh, S, hs) x (B x nh, hs, S) -> (B x nh, S, S)
        y = scaled_dot_product_attention(
            q=q, k=k, v=v, att_mask=att_mask, dropout=self.attn_drop, causal=self.causal
        )
        return y



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS: This comes almost as-is from the Triton dropout tutorial
# https://raw.githubusercontent.com/openai/triton/master/python/tutorials/04-low-memory-dropout.py

import triton
import triton.language as tl


# fmt: off
@triton.autotune(
    configs=[
        triton.Config({"BLOCK_SIZE" : 256}, num_warps=1),
        triton.Config({"BLOCK_SIZE" : 512}, num_warps=2),
        triton.Config({"BLOCK_SIZE" : 1024}, num_warps=4),
        triton.Config({"BLOCK_SIZE" : 2048}, num_warps=8),
        triton.Config({"BLOCK_SIZE" : 4096}, num_warps=8),
    ],
    key=["N"],
)
@triton.jit
def k_dropout(
    Y, X, SEEDS,
    stride,
    N,
    p,
    **meta,
):
    """
    Apply dropout on an input tensor
    Y : Output (M, N)
    X : Input (M, N)
    S : Seeds (M,)
    p : dropout probability
    """
    # fmt: on

    # compute memory offsets of elements handled by this instance
    BLOCK_SIZE = meta["BLOCK_SIZE"]
    row = tl.program_id(axis=0)
    col = tl.program_id(axis=1)
    offsets = row * stride + col * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = col * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE) < N

    # load data from x
    x_ptrs = X + offsets
    x = tl.load(x_ptrs, mask=mask)

    # randomly prune it
    seed = SEEDS + row
    random = tl.rand(seed.to(tl.int32), offsets)
    x_keep = random > p

    # write-back
    zero = 0.
    zero = zero.to(x.dtype)
    output = tl.where(x_keep, (x / (1 - p)).to(x.dtype), zero)
    y_ptrs = Y + offsets
    tl.store(y_ptrs, output, mask=mask)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch
import triton
import triton.language as tl

from xformers.triton.configs_matmul import kernel_config

# CREDITS: Initially inspired by the Triton tutorial on matrix multiplications


# fmt: off
@triton.autotune(
    configs=kernel_config,
    key=["M", "N", "K"],
)
@triton.jit
def kernel_fma(
    # Pointers to matrices
    D, A, W, C, In,
    # Matrix dimensions
    M, N, K,
    # The stride variables represent how much to increase the ptr by when moving by 1
    # element in a particular dimension. E.g. stride_am is how much to increase a_ptr
    # by to get the element one row down (A has M rows)
    stride_db,  stride_dm, stride_dn,
    stride_ab, stride_am, stride_ak,
    stride_wn, stride_wk,
    stride_ib,  stride_im, stride_in,
    # Meta-parameters
    **META,
):
    # fmt: on

    """
    Kernel for computing D = activation(A x W + C)

    - A has shape (B, M, K)
    - W has shape (K, N)
    - C has shape (N,)
    - D has shape (B, M, N)
    - In (optional) has shape (B, M, N)

    'In' optionally saves the A x W + C intermediate for backward computations
    """

    # extract metaparameters
    BLOCK_M, GROUP_M = META["BLOCK_ROW"], META["GROUP_M"]
    BLOCK_N, BLOCK_K = META["BLOCK_COL"], META["BLOCK_K"]

    # programs are grouped together to improve L2 hit rate
    pid, batch_id = tl.program_id(axis=0), tl.program_id(axis=1)

    num_pid_m = tl.cdiv(M, BLOCK_M)  # number of program ids along the M axis
    num_pid_n = tl.cdiv(N, BLOCK_N)  # number of programs ids along the N axis
    num_pid_in_group = GROUP_M * num_pid_n  # number of programs in group
    group_id = pid // num_pid_in_group  # id of the group this program is in
    first_pid_m = group_id * GROUP_M  # row-id of the first program in the group
    GROUP_M = min(
        num_pid_m - first_pid_m, GROUP_M
    )  # if `num_pid_m` isn't divisible by `GROUP_M`, the last group is smaller

    # *within groups*, programs are ordered in a column-major order
    # row-id of the program in the *launch grid*
    pid_m = first_pid_m + (pid % GROUP_M)

    # col-id of the program in the *launch grid*
    pid_n = (pid % num_pid_in_group) // GROUP_M

    # rm (resp. rn) denotes a range of indices
    # for rows (resp. col) of C
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)

    # rk denotes a range of indices for columns
    # (resp. rows) of A (resp. B)
    rk = tl.arange(0, BLOCK_K)

    # the memory addresses of elements in the first block of
    # A and W can be computed using numpy-style broadcasting
    D += rm[:, None] * stride_dm + rn[None, :] * stride_dn + batch_id * stride_db
    A += rm[:, None] * stride_am + rk[None, :] * stride_ak + batch_id * stride_ab
    W += rk[:, None] * stride_wk + rn[None, :] * stride_wn

    # initialize and iteratively update accumulator
    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)

    if META["BIAS"]:
        bias = tl.load(C + rn, mask=rn < N, other=0.0).to(tl.float32)
        acc += bias[None, :]

    for _ in range(K, 0, -BLOCK_K):
        # block level matrix multiplication
        a = tl.load(A)
        w = tl.load(W)
        acc += tl.dot(a, w).to(tl.float32)

        # increment pointers so that the next blocks of A and B
        # are loaded during the next iteration
        A += BLOCK_K * stride_ak
        W += BLOCK_K * stride_wk

    # optional: save the activation inputs
    if META["SAVE_ACT_INPUTS"]:
        In += rm[:, None] * stride_im + rn[None, :] * stride_in + batch_id * stride_ib
        tl.store(In, acc, mask=(rm[:, None] < M) & (rn[None, :] < N))

    # optional: fused activation (while the data is in shared memory)
    if META["ACTIVATION"]:
        acc = META["ACTIVATION"](acc)

    # write back result
    tl.store(D, acc, mask=(rm[:, None] < M) & (rn[None, :] < N))


# Activation needs to be a triton kernel
def fused_matmul(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: Optional[torch.Tensor],
    activation=None,
    save_inputs: bool = False
):
    """
    Compute e = activation(x @ weight + bias).
    This wrapper kicks the `kernel_fma` Triton kernel
    """

    if x.ndim == 2:
        x = x.unsqueeze(0)
        _should_squeeze = True
    else:
        _should_squeeze = False

    assert (
        x.shape[2] == weight.shape[1]
    ), f"Incompatible dimensions in between inputs and weight, {x.shape} - {weight.shape}"
    assert bias is None or bias.is_contiguous()
    assert (
        bias is None or bias.shape[0] == weight.shape[0]
    ), "Incompatible dimensions in between weight and bias"

    B, M, K = x.shape
    N, K = weight.shape

    # FIXME: @lefaudeux
    assert (
        K % 32 == 0
    ), "We don't check memory-out-of-bounds with K so K must be divisible by BLOCK_K"

    outputs = torch.empty((B, M, N), device=x.device, dtype=x.dtype)
    act_inputs = torch.empty_like(outputs) if save_inputs else outputs

    # 1D launch kernel where each block gets its own program.
    def grid(META):
        return (
            triton.cdiv(M, META["BLOCK_ROW"]) * triton.cdiv(N, META["BLOCK_COL"]),
            B,
        )

    # fmt: off
    kernel_fma[grid](
        # data ptrs
        outputs, x, weight,
        bias if bias is not None else x,  # auto skip bias if not present
        act_inputs,
        # shapes
        M, N, K,
        # strides
        outputs.stride(0), outputs.stride(1), outputs.stride(2),
        x.stride(0), x.stride(1), x.stride(2),
        weight.stride(0), weight.stride(1),
        act_inputs.stride(0), act_inputs.stride(1), act_inputs.stride(2),
        # optional fused activation
        ACTIVATION=activation,
        # optional fused bias
        BIAS=bias is not None,
        # speed optimization: group the programs
        # improve on data reuse in L2 cache
        GROUP_M=8,
        BLOCK_K=32,
        SAVE_ACT_INPUTS=save_inputs
    )
    # fmt: on

    if _should_squeeze:
        outputs.squeeze_()

    return (outputs, act_inputs) if save_inputs else (outputs, None)


# fmt: off
@triton.autotune(
    configs=kernel_config,
    key=["M", "N", "K"],
)
@triton.jit
def kernel_fma_grad_in(
    # Pointers to all the tensors
    GRAD_IN, GRAD_ACT, ACT_IN, GRAD_OUT, W,
    # Tensor dimensions
    M, N, K,
    # strides for all the gradients
    stride_gib, stride_gim, stride_gik,
    stride_gab, stride_gam, stride_gan,
    stride_gob, stride_gom, stride_gon,
    # strides for the extra data
    stride_aib, stride_aim, stride_ain,
    stride_wn, stride_wk,
    # Meta-parameters
    **META,
):
    # fmt: on
    """
    Kernel for computing `grad_out = grad_in * activation_grad(inputs) @ W^T`
    - grad_out has shape (B, M, N)
    - W has shape (K, N)
    - grad_in has shape (B, M, K)
    - X has shape (B, M, K)
    """
    # extract metaparameters
    BLOCK_M, GROUP_M = META["BLOCK_ROW"], META["GROUP_M"]
    BLOCK_N, BLOCK_K = META["BLOCK_N"], META["BLOCK_COL"]

    # programs are grouped together to improve L2 hit rate
    pid, batch_id = tl.program_id(axis=0), tl.program_id(axis=1)

    num_pid_m = tl.cdiv(M, BLOCK_M)  # number of program ids along the M axis
    num_pid_k = tl.cdiv(K, BLOCK_K)  # number of programs ids along the N axis
    num_pid_in_group = GROUP_M * num_pid_k  # number of programs in group
    group_id = pid // num_pid_in_group  # id of the group this program is in
    first_pid_m = group_id * GROUP_M  # row-id of the first program in the group
    GROUP_M = min(
        num_pid_m - first_pid_m, GROUP_M
    )  # if `num_pid_m` isn't divisible by `GROUP_M`, the last group is smaller

    # *within groups*, programs are ordered in a column-major order
    pid_m = first_pid_m + (pid % GROUP_M)  # row-id of the program in the *launch grid*
    pid_k = (
        pid % num_pid_in_group
    ) // GROUP_M  # col-id of the program in the *launch grid*

    # memory ranges
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rk = pid_k * BLOCK_K + tl.arange(0, BLOCK_K)
    rn = tl.arange(0, BLOCK_N)

    # memory blocks can be computed using numpy-style broadcasting
    GRAD_OUT += rm[:, None] * stride_gom + rn[None, :] * stride_gon + batch_id * stride_gob
    GRAD_ACT += rm[:, None] * stride_gam + rn[None, :] * stride_gan + batch_id * stride_gab
    ACT_IN += rm[:, None] * stride_aim + rn[None, :] * stride_ain + batch_id * stride_aib
    GRAD_IN += rm[:, None] * stride_gim + rk[None, :] * stride_gik + batch_id * stride_gib
    W += rn[:, None] * stride_wn + rk[None, :] * stride_wk

    # initialize and iteratively update accumulator
    grad_in = tl.zeros((BLOCK_M, BLOCK_K), dtype=tl.float32)
    act_grad_fn = META["ACTIVATION_GRAD"]

    for _ in range(N, 0, -BLOCK_N):
        grad_out = tl.load(GRAD_OUT)  # BLOCK_M x BLOCK_N
        w = tl.load(W)  # BLOCK_N x BLOCK_K

        # optional fused activation gradient (while the data is in shared memory)
        if META["ACTIVATION_GRAD"]:
            if META["ACTIVATION_GRAD_REQ_INPUTS"]:
                # This activation requires its inputs
                act_input = tl.load(ACT_IN)
                grad_act = act_grad_fn(act_input)
                ACT_IN += BLOCK_N * stride_ain
            else:
                # Save some time, we can reuse the outputs to know about the grad
                grad_act = act_grad_fn(grad_out)

            grad_out *= grad_act.to(grad_out.dtype)

        # store grad_act as an intermediate, will be used for grad/weight and grad/bias
        if META["SAVE_ACT_GRAD"]:
            tl.store(GRAD_ACT, grad_out)

        # gradient #1: input with respect to outputs
        # grad_in is grad_out scaled by the (transposed) weight
        grad_in += tl.dot(grad_out, w)

        # increment pointers so that the next blocks of A and B are loaded during the next iteration
        GRAD_OUT += BLOCK_N * stride_gon
        GRAD_ACT += BLOCK_N * stride_gan
        W += BLOCK_N * stride_wn

    # write back result
    tl.store(GRAD_IN, grad_in, mask=(rm[:, None] < M) & (rk[None, :] < K))  # type promotion or downgrade is automatic


# Activation needs to be a triton kernel
def fused_matmul_backward(
    grad_out: torch.Tensor,
    inputs: torch.Tensor,
    weight: torch.Tensor,
    trainable_weight: bool,
    trainable_bias: bool,
    activation_inputs: Optional[torch.Tensor],
    activation_grad=None,
    activation_grad_req_inputs: bool = False,
):
    """
    Compute grad_in = activation^-1(grad_out) @ weight.transpose()

    .. note: The weight buffer is transposed on the fly
    """

    if grad_out.ndim == 2:
        # Add the batch dimension
        # This is inelegant and maybe slow, but never really used in the xformers context
        grad_out = grad_out.unsqueeze(0)
        _should_squeeze = True
    else:
        _should_squeeze = False

    assert (
        grad_out.shape[2] == weight.shape[0]
    ), "Incompatible dimensions in between grad_out and weight"

    B, M, N = grad_out.shape
    N, K = weight.shape

    grad_in = torch.empty((B, M, K), device=grad_out.device, dtype=grad_out.dtype)
    grad_act = torch.empty_like(grad_out)

    # Compute the gradient for the inputs
    def grid(META):
        return (
            triton.cdiv(M, META["BLOCK_ROW"]) * triton.cdiv(K, META["BLOCK_COL"]),
            B,
        )

    if activation_inputs is None:
        # place holder, this will not be used really
        activation_inputs = grad_out

    # fmt: off
    kernel_fma_grad_in[grid](
        # data ptrs
        grad_in, grad_act, activation_inputs, grad_out, weight,
        # shapes
        M, N, K,
        # strides
        grad_in.stride(0), grad_in.stride(1), grad_in.stride(2),
        grad_act.stride(0), grad_act.stride(1), grad_act.stride(2),
        grad_out.stride(0), grad_out.stride(1), grad_out.stride(2),
        activation_inputs.stride(0), activation_inputs.stride(1), activation_inputs.stride(2),
        weight.stride(0), weight.stride(1),
        # optional fused activation
        ACTIVATION_GRAD=activation_grad,
        # data reuse optimization
        GROUP_M=16,
        BLOCK_N=32,
        ACTIVATION_GRAD_REQ_INPUTS=activation_grad_req_inputs,
        SAVE_ACT_GRAD=trainable_weight or trainable_bias
    )
    # fmt: on

    grad_bias = torch.sum(grad_act, dim=[0, 1]) if trainable_bias else None

    # Reuse Triton optimized matmul
    grad_weight = None
    if trainable_weight:
        grad_act_ = torch.reshape(grad_act, (grad_act.shape[0]*grad_act.shape[1], grad_act.shape[2])).transpose(1, 0)
        if inputs.ndim == 2:
            inputs = inputs.unsqueeze(0)
        inputs_ = torch.reshape(inputs, (inputs.shape[0]*inputs.shape[1], inputs.shape[2]))
        grad_weight = triton.ops.matmul(grad_act_, inputs_)

    if _should_squeeze:
        grad_in = grad_in.squeeze_()

    del grad_act

    return grad_in, grad_weight, grad_bias



# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Main training script for the image classification task."""
import functools
import itertools
import json
import os
import time

from absl import app
from absl import flags
from absl import logging
from flax import jax_utils
from flax import nn
from flax import optim
from flax.metrics import tensorboard
from flax.training import checkpoints
from flax.training import common_utils
import jax
from jax import random
import jax.nn
import jax.numpy as jnp
from lra_benchmarks.image import task_registry
from lra_benchmarks.models.transformer import transformer
from lra_benchmarks.utils import train_utils
from ml_collections import config_flags
import tensorflow.compat.v2 as tf

FLAGS = flags.FLAGS

config_flags.DEFINE_config_file(
    'config', None, 'Training configuration.', lock_config=True)
flags.DEFINE_string(
    'model_dir', default=None, help='Directory to store model data.')
flags.DEFINE_string('task_name', default='mnist', help='Name of the task')
flags.DEFINE_bool(
    'eval_only', default=False, help='Run the evaluation on the test data.')


def create_model(key, flax_module, input_shape, model_kwargs):
  """Creates and initializes the model."""

  @functools.partial(jax.jit, backend='cpu')
  def _create_model(key):
    module = flax_module.partial(**model_kwargs)
    with nn.stateful() as init_state:
      with nn.stochastic(key):
        _, initial_params = module.init_by_shape(key,
                                                 [(input_shape, jnp.float32)])
        model = nn.Model(module, initial_params)
    return model, init_state

  return _create_model(key)


def create_optimizer(model, learning_rate, weight_decay):
  optimizer_def = optim.Adam(
      learning_rate, beta1=0.9, beta2=0.98, eps=1e-9, weight_decay=weight_decay)
  optimizer = optimizer_def.create(model)
  return optimizer


def compute_metrics(logits, labels, num_classes, weights):
  """Compute summary metrics."""
  loss, weight_sum = train_utils.compute_weighted_cross_entropy(
      logits, labels, num_classes, weights=weights)
  acc, _ = train_utils.compute_weighted_accuracy(logits, labels, weights)
  metrics = {
      'loss': loss,
      'accuracy': acc,
      'denominator': weight_sum,
  }
  metrics = jax.lax.psum(metrics, 'batch')
  return metrics


def get_model(init_rng, input_shape, model_type, model_kwargs):
  """Create and initialize the model.

  Args:
    init_rng: float; Jax PRNG key.
    input_shape: tuple; Tuple indicating input shape.
    model_type: str; Type of Transformer model to create.
    model_kwargs: keyword argument to the model.

  Returns:
    Initialized model.
  """
  if model_type == 'transformer':
    return create_model(init_rng, transformer.TransformerEncoder, input_shape,
                        model_kwargs)
  else:
    raise ValueError('Model type not supported')


def train_step(optimizer,
               state,
               batch,
               learning_rate_fn,
               num_classes,
               flatten_input=True,
               grad_clip_norm=None,
               dropout_rng=None):
  """Perform a single training step."""
  train_keys = ['inputs', 'targets']
  (inputs, targets) = [batch.get(k, None) for k in train_keys]
  if flatten_input:
    inputs = inputs.reshape(inputs.shape[0], -1)

  # We handle PRNG splitting inside the top pmap, rather
  # than handling it outside in the training loop - doing the
  # latter can add some stalls to the devices.
  dropout_rng, new_dropout_rng = random.split(dropout_rng)

  def loss_fn(model):
    """Loss function used for training."""
    with nn.stateful(state) as new_state:
      with nn.stochastic(dropout_rng):
        logits = model(inputs, train=True)
    loss, weight_sum = train_utils.compute_weighted_cross_entropy(
        logits, targets, num_classes=num_classes, weights=None)
    mean_loss = loss / weight_sum
    return mean_loss, (new_state, logits)

  step = optimizer.state.step
  lr = learning_rate_fn(step)
  grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
  (_, (new_state, logits)), grad = grad_fn(optimizer.target)
  grad = jax.lax.pmean(grad, 'batch')
  if grad_clip_norm:
    # Optionally resize the global gradient to a maximum norm.
    gradients, _ = jax.tree_flatten(grad)
    g_l2 = jnp.sqrt(sum([jnp.vdot(p, p) for p in gradients]))
    g_factor = jnp.minimum(1.0, grad_clip_norm / g_l2)
    grad = jax.tree_map(lambda p: g_factor * p, grad)
  new_optimizer = optimizer.apply_gradient(grad, learning_rate=lr)
  metrics = compute_metrics(logits, targets, num_classes, weights=None)
  metrics['learning_rate'] = lr
  return new_optimizer, new_state, metrics, new_dropout_rng


def eval_step(model, state, batch, num_classes, flatten_input=True):
  eval_keys = ['inputs', 'targets']
  (inputs, targets) = [batch.get(k, None) for k in eval_keys]
  if flatten_input:
    inputs = inputs.reshape(inputs.shape[0], -1)
  if jax.tree_leaves(state):
    state = jax.lax.pmean(state, 'batch')
  with nn.stateful(state, mutable=False):
    logits = model(inputs, train=False)
  return compute_metrics(logits, targets, num_classes, weights=None)


def test(optimizer, state, p_eval_step, step, test_ds, summary_writer,
         model_dir):
  """Test the flax module in optimizer on test_ds.

  Args:
    optimizer: flax optimizer (contains flax module).
    state: model state, e.g. batch statistics.
    p_eval_step: fn; Pmapped evaluation step function.
    step: int; Number of training steps passed so far.
    test_ds: tf.dataset; Test dataset.
    summary_writer: tensorflow summary writer.
    model_dir: model directory.
  """
  # Test Metrics
  test_metrics = []
  test_iter = iter(test_ds)
  for _, test_batch in zip(itertools.repeat(1), test_iter):
    # pylint: disable=protected-access
    test_batch = common_utils.shard(
        jax.tree_map(lambda x: x._numpy(), test_batch))
    # pylint: enable=protected-access
    metrics = p_eval_step(optimizer.target, state, test_batch)
    test_metrics.append(metrics)
  test_metrics = common_utils.get_metrics(test_metrics)
  test_metrics_sums = jax.tree_map(jnp.sum, test_metrics)
  test_denominator = test_metrics_sums.pop('denominator')
  test_summary = jax.tree_map(
      lambda x: x / test_denominator,  # pylint: disable=cell-var-from-loop
      test_metrics_sums)
  logging.info('test in step: %d, loss: %.4f, acc: %.4f', step,
               test_summary['loss'], test_summary['accuracy'])
  if jax.host_id() == 0:
    for key, val in test_summary.items():
      summary_writer.scalar(f'test_{key}', val, step)
    summary_writer.flush()
  with tf.io.gfile.GFile(os.path.join(model_dir, 'results.json'), 'w') as f:
    json.dump(jax.tree_map(lambda x: x.tolist(), test_summary), f)


def train_loop(config, dropout_rngs, eval_ds, eval_freq, num_eval_steps,
               num_train_steps, optimizer, state, p_eval_step, p_train_step,
               start_step, train_iter, summary_writer):
  """Training loop.

  Args:
    config: experiment config.
    dropout_rngs: float array; Jax PRNG key.
    eval_ds: tf.dataset; Evaluation dataset.
    eval_freq: int; Evaluation frequency;
    num_eval_steps: int; Number of evaluation steps.
    num_train_steps: int; Number of training steps.
    optimizer: flax optimizer.
    state: model state, e.g. batch statistics.
    p_eval_step: fn; Pmapped evaluation step function.
    p_train_step: fn; Pmapped train step function.
    start_step: int; global training step.
    train_iter: iter(tf.dataset); Training data iterator.
    summary_writer: tensorflow summary writer.

  Returns:
    optimizer, global training step
  """
  metrics_all = []
  tick = time.time()
  logging.info('Starting training')
  logging.info('====================')

  step = 0
  for step, batch in zip(range(start_step, num_train_steps), train_iter):
    batch = common_utils.shard(jax.tree_map(lambda x: x._numpy(), batch))  # pylint: disable=protected-access
    optimizer, state, metrics, dropout_rngs = p_train_step(
        optimizer, state, batch, dropout_rng=dropout_rngs)
    metrics_all.append(metrics)
    # Save a Checkpoint
    if ((step % config.checkpoint_freq == 0 and step > 0) or
        step == num_train_steps - 1):
      if jax.host_id() == 0 and config.save_checkpoints:
        # Save unreplicated optimizer + model state.
        checkpoints.save_checkpoint(
            FLAGS.model_dir,
            (jax_utils.unreplicate(optimizer), jax_utils.unreplicate(state)),
            step)

    # Periodic metric handling.
    if step % eval_freq == 0 and step > 0:
      metrics_all = common_utils.get_metrics(metrics_all)
      lr = metrics_all.pop('learning_rate').mean()
      metrics_sums = jax.tree_map(jnp.sum, metrics_all)
      denominator = metrics_sums.pop('denominator')
      summary = jax.tree_map(lambda x: x / denominator, metrics_sums)  # pylint: disable=cell-var-from-loop
      summary['learning_rate'] = lr
      # Calculate (clipped) perplexity after averaging log-perplexities:
      logging.info('train in step: %d, loss: %.4f, acc: %.4f', step,
                   summary['loss'], summary['accuracy'])
      if jax.host_id() == 0:
        tock = time.time()
        steps_per_sec = eval_freq / (tock - tick)
        tick = tock
        summary_writer.scalar('examples_per_second',
                              steps_per_sec * config.batch_size, step)
        for key, val in summary.items():
          summary_writer.scalar(f'train_{key}', val, step)
        summary_writer.flush()
      # Reset metric accumulation for next evaluation cycle.
      metrics_all = []

      # Eval Metrics
      eval_metrics = []
      eval_iter = iter(eval_ds)
      if num_eval_steps == -1:
        num_iter = itertools.repeat(1)
      else:
        num_iter = range(num_eval_steps)
      for _, eval_batch in zip(num_iter, eval_iter):
        # pylint: disable=protected-access
        eval_batch = common_utils.shard(
            jax.tree_map(lambda x: x._numpy(), eval_batch))
        # pylint: enable=protected-access
        metrics = p_eval_step(optimizer.target, state, eval_batch)
        eval_metrics.append(metrics)
      eval_metrics = common_utils.get_metrics(eval_metrics)
      eval_metrics_sums = jax.tree_map(jnp.sum, eval_metrics)
      eval_denominator = eval_metrics_sums.pop('denominator')
      eval_summary = jax.tree_map(
          lambda x: x / eval_denominator,  # pylint: disable=cell-var-from-loop
          eval_metrics_sums)
      logging.info('eval in step: %d, loss: %.4f, acc: %.4f', step,
                   eval_summary['loss'], eval_summary['accuracy'])
      if jax.host_id() == 0:
        for key, val in eval_summary.items():
          summary_writer.scalar(f'val_{key}', val, step)
        summary_writer.flush()
  return optimizer, state, step


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  tf.enable_v2_behavior()

  config = FLAGS.config
  logging.info('===========Config Dict============')
  logging.info(config)
  batch_size = config.batch_size
  learning_rate = config.learning_rate
  num_train_steps = config.num_train_steps
  num_eval_steps = config.num_eval_steps
  eval_freq = config.eval_frequency
  random_seed = config.random_seed
  model_type = config.model_type

  if jax.host_id() == 0:
    summary_writer = tensorboard.SummaryWriter(
        os.path.join(FLAGS.model_dir, 'summary'))
  else:
    summary_writer = None

  if batch_size % jax.device_count() > 0:
    raise ValueError('Batch size must be divisible by the number of devices')

  logging.info('Training on %s', FLAGS.task_name)

  if model_type in ['wideresnet', 'resnet', 'simple_cnn']:
    normalize = True
  else:  # transformer-based models
    normalize = False
  (train_ds, eval_ds, test_ds, num_classes, vocab_size,
   input_shape) = task_registry.TASK_DATA_DICT[FLAGS.task_name](
       n_devices=jax.local_device_count(),
       batch_size=batch_size,
       normalize=normalize)
  train_iter = iter(train_ds)
  model_kwargs = {}
  flatten_input = True

  if model_type in ['wideresnet', 'resnet', 'simple_cnn']:
    model_kwargs.update({
        'num_classes': num_classes,
    })
    flatten_input = False

  else:  # transformer models
    # we will flatten the input
    bs, h, w, c = input_shape
    assert c == 1
    input_shape = (bs, h * w * c)
    model_kwargs.update({
        'vocab_size': vocab_size,
        'max_len': input_shape[1],
        'classifier': True,
        'num_classes': num_classes,
    })

  model_kwargs.update(config.model)

  rng = random.PRNGKey(random_seed)
  rng = jax.random.fold_in(rng, jax.host_id())
  rng, init_rng = random.split(rng)
  # We init the first set of dropout PRNG keys, but update it afterwards inside
  # the main pmap'd training update for performance.
  dropout_rngs = random.split(rng, jax.local_device_count())

  model, state = get_model(init_rng, input_shape, model_type, model_kwargs)

  optimizer = create_optimizer(model, learning_rate, config.weight_decay)
  del model  # Don't keep a copy of the initial model.

  start_step = 0
  if config.restore_checkpoints:
    # Restore unreplicated optimizer + model state from last checkpoint.
    optimizer, state = checkpoints.restore_checkpoint(FLAGS.model_dir,
                                                      (optimizer, state))
    # Grab last step.
    start_step = int(optimizer.state.step)

  # Replicate optimizer and state
  optimizer = jax_utils.replicate(optimizer)
  state = jax_utils.replicate(state)

  learning_rate_fn = train_utils.create_learning_rate_scheduler(
      factors=config.factors,
      base_learning_rate=learning_rate,
      warmup_steps=config.warmup,
      steps_per_cycle=config.get('steps_per_cycle', None),
  )
  p_train_step = jax.pmap(
      functools.partial(
          train_step,
          learning_rate_fn=learning_rate_fn,
          num_classes=num_classes,
          grad_clip_norm=config.get('grad_clip_norm', None),
          flatten_input=flatten_input),
      axis_name='batch')

  p_eval_step = jax.pmap(
      functools.partial(
          eval_step, num_classes=num_classes, flatten_input=flatten_input),
      axis_name='batch',
  )

  optimizer, state, step = train_loop(config, dropout_rngs, eval_ds, eval_freq,
                                      num_eval_steps, num_train_steps,
                                      optimizer, state, p_eval_step,
                                      p_train_step, start_step, train_iter,
                                      summary_writer)

  logging.info('Starting testing')
  logging.info('====================')
  test(optimizer, state, p_eval_step, step, test_ds, summary_writer,
       FLAGS.model_dir)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration and hyperparameter sweeps."""

from lra_benchmarks.image.configs.pathfinder128 import base_pathfinder128_config


def get_config():
  """Get the hyperparameter configuration."""
  config = base_pathfinder128_config.get_config()
  config.model_type = "transformer"
  return config


def get_hyper(hyper):
  return hyper.product([])



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import torch

from xformers.components import reversible as rv
from xformers.factory.block_factory import (
    BlockType,
    xFormerDecoderBlock,
    xFormerDecoderConfig,
    xFormerEncoderBlock,
    xFormerEncoderConfig,
)


@dataclass(init=False)
class xFormerStackConfig:
    """
    A stack is defined by the definition of a given block, and an optional repetition factor
    """

    block_config: Union[xFormerEncoderConfig, xFormerDecoderConfig]
    num_layers: int
    reversible: bool

    def __init__(
        self, block_config: Dict[str, Any], reversible: bool = False, *args, **kwargs
    ):
        if block_config["block_type"] == BlockType.Encoder:
            self.block_config = xFormerEncoderConfig(**block_config)
        else:
            self.block_config = xFormerDecoderConfig(**block_config)

        # Convenience: make num_layers optional, so that a stack at that point could
        # only be defined by a given block, and no repetition
        if "num_layers" in block_config.keys():
            self.num_layers = block_config["num_layers"]
        else:
            self.num_layers = 1

        self.reversible = reversible


@dataclass(init=False)
class xFormerConfig:
    stack_configs: List[xFormerStackConfig]

    def __init__(self, stack_configs: List[Dict[str, Any]]):
        # Type all the configurations. Possible typos are caught here
        self.stack_configs = [xFormerStackConfig(**config) for config in stack_configs]

        # Check that the reversible setting is not alternating, which
        # - makes little sense, since you loose all the reversible benefits
        # - may break
        # Reversible is only allowed on the encoder side

        reversible = [
            c.reversible
            for c in filter(
                lambda x: x.block_config.block_type == "encoder", self.stack_configs
            )
        ]
        non_reversible = [not rev for rev in reversible]

        try:
            assert all(reversible) or all(non_reversible), (
                "All layers need to have the same reversibility setting. "
                + f"Currently {reversible}"
            )
        except AssertionError:
            print(self.stack_configs)
            print(reversible)
            assert False


class xFormer(torch.nn.Module):
    def __init__(
        self, stack_configs: Union[xFormerStackConfig, List[xFormerStackConfig]]
    ):
        """
        Given a serialized configuration, generate the corresponding model.
        This is only a helper and can easily be bypassed
        """

        super().__init__()

        encoders: List[torch.nn.Module] = []
        decoders: List[torch.nn.Module] = []

        self.reversible_encoder = False
        self.enc_pose_encoding = None
        self.dec_pose_encoding = None

        # Convenience, users can pass either a list of configs or a single one
        if not isinstance(stack_configs, List):
            stack_configs = [stack_configs]

        # Unroll the configs and build the model
        for stack in stack_configs:
            config = stack.block_config

            # Handle either Encoder or Decoder stacks
            builder = (
                xFormerEncoderBlock.from_config
                if isinstance(config, xFormerEncoderConfig)
                else xFormerDecoderBlock.from_config
            )
            recipient = (
                encoders if isinstance(config, xFormerEncoderConfig) else decoders
            )

            # Build up the stack
            for i in range(stack.num_layers):
                # Label where this layer is in the stack
                # (for instance useful for the positional encoding, or late layer norm)
                if i > 0:
                    config.layer_position.mark_not_first()
                if i < stack.num_layers - 1:
                    config.layer_position.mark_not_last()
                block = builder(config)  # type: ignore

                # If reversible: extract the reversible sub-parts, else append the block as-is
                if stack.reversible:
                    # WARNING: only one pose encoding is saved here (not Focal Transformer compatible for instance)
                    assert isinstance(config, xFormerEncoderConfig)
                    if block.pose_encoding is not None:
                        self.enc_pose_encoding = block.pose_encoding
                    self.reversible_encoder = True

                    f, g = xFormerEncoderBlock.get_reversible_layer(config)
                    recipient.append(torch.nn.ModuleList([f, g]))
                else:
                    recipient.append(block)  # type: ignore

        self.encoders: torch.nn.Module = (
            rv.ReversibleSequence(torch.nn.ModuleList(encoders))
            if self.reversible_encoder
            else torch.nn.ModuleList(encoders)
        )
        self.decoders = torch.nn.ModuleList(decoders)

        if len(self.decoders) > 0:
            # Use Xavier init for encoding/decoding tasks
            self._reset_parameters()

    @classmethod
    def from_config(cls, config: xFormerConfig):
        return cls(config.stack_configs)

    def _reset_parameters(self):
        r"""Initiate parameters in the transformer model
        following the Xavier distribution."""

        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)

    def forward(
        self,
        src: torch.Tensor,
        tgt: Optional[torch.Tensor] = None,
        encoder_input_mask: Optional[torch.Tensor] = None,
        decoder_input_mask: Optional[torch.Tensor] = None,
    ) -> Optional[torch.Tensor]:

        # Encode to latent space if encoder is present
        if len(list(self.encoders.parameters())) > 0:
            encoders = self.encoders
            if isinstance(encoders, torch.nn.ModuleList):
                memory = src.clone()
                for encoder in encoders:
                    memory = encoder(memory, input_mask=encoder_input_mask)
            else:
                if self.enc_pose_encoding:
                    memory = self.enc_pose_encoding(src)

                # pyre-fixme[61]: `memory` is not always initialized here.
                # Reversible Encoder
                x = torch.cat([memory, memory], dim=-1)

                # TODO: pass in key and value independently.
                kwargs = {"att_mask": encoder_input_mask}
                x = encoders(x, **kwargs)
                memory = torch.stack(x.chunk(2, dim=-1)).mean(dim=0)

            if not self.decoders:
                return memory

        # If decoder: either use the encoder ouput, or just decode, both options are possible
        if len(self.decoders) > 0:
            tgt = src.clone() if tgt is None else tgt

            for decoder in self.decoders:
                tgt = decoder(
                    target=tgt,
                    # pyre-fixme[61]: `memory` is not always initialized here.
                    memory=memory,
                    input_mask=decoder_input_mask,
                )

            return tgt

        return None



# Copyright 2021 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mapping tasks to data loaders."""

import functools
from lra_benchmarks.image import input_pipeline

TASK_DATA_DICT = {
    'cifar10':
        input_pipeline.get_cifar10_datasets,
    'pathfinder32_easy':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=32,
            split='easy'),
    'pathfinder32_inter':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=32,
            split='intermediate'),
    'pathfinder32_hard':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=32,
            split='hard'),
    'pathfinder64_easy':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=64,
            split='easy'),
    'pathfinder64_inter':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=64,
            split='intermediate'),
    'pathfinder64_hard':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=64,
            split='hard'),
    'pathfinder128_easy':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=128,
            split='easy'),
    'pathfinder128_inter':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=128,
            split='intermediate'),
    'pathfinder128_hard':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=128,
            split='hard'),
    'pathfinder256_easy':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=256,
            split='easy'),
    'pathfinder256_inter':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=256,
            split='intermediate'),
    'pathfinder256_hard':
        functools.partial(
            input_pipeline.get_pathfinder_base_datasets,
            resolution=256,
            split='hard'),
}
